/*

Theremincampoop, a webcam-based instrument
  version 1.0, October 25th, 2008

  Copyright (C) 2008 Jari Komppa

  This software is provided 'as-is', without any express or implied
  warranty.  In no event will the authors be held liable for any damages
  arising from the use of this software.

  Permission is granted to anyone to use this software for any purpose,
  including commercial applications, and to alter it and redistribute it
  freely, subject to the following restrictions:

  1. The origin of this software must not be misrepresented; you must not
     claim that you wrote the original software. If you use this software
     in a product, an acknowledgment in the product documentation would be
     appreciated but is not required.
  2. Altered source versions must be plainly marked as such, and must not be
     misrepresented as being the original software.
  3. This notice may not be removed or altered from any source distribution.

  http://iki.fi/sol

*/

#include <windows.h>
#include <stdlib.h>
#include <math.h>
#include "SDL.h"
#include "escapi.h"

#define MAX(a,b) ((a) > (b))?(a):(b)
#define MIN(a,b) ((a) < (b))?(a):(b)

// some options..

//#define FIGHT_AUTO_LIGHT
//#define BOXFILTER


SDL_Surface *gScreen;
struct SimpleCapParams gCapture;

int minval = 0;
int scaler = 0xff;
int scaleg = 0xff;
int scaleb = 0xff;
int vbuf[320*240] = {0};
int level_r=0;
int level_l=0;

int audiolevels[65536];

int v(int x, int y, int p)
{
    if (x < 0) x = 0;
    if (x > 319) x = 319;
    if (y < 0) y = 0;
    if (y > 239) y = 239;
    int val = ((gCapture.mTargetBuf[y*320+x] - minval) >> p) & 0xff;
    switch (p)
    {
    case  0: val = (val * scaleb) >> 8; break;
    case  8: val = (val * scaleg) >> 8; break;
    default: val = (val * scaler) >> 8; break;
    }
    return val;
}


// we don't care about hue, so this simplified a bit.. =) (SV from HSV)
// based on cgpp:foley, van dam etc
int rgb2sv(int r, int g, int b) 
{
    float fr, fg, fb;
	float fs, fv;
	float max, min, diff;

    fr = (float)r / 0xff;
    fg = (float)g / 0xff;
    fb = (float)b / 0xff;

	max = MAX(MAX(fr, fg), fb);
	min = MIN(MIN(fr, fg), fb);
	diff = max - min;

	if (max == 0)
    {
		return 0;
    }
	else 
    {
		fv = max;
		fs = diff/max;
	}

    return (((int)floor(fs * 0xff)) << 8) |
           (((int)floor(fv * 0xff)) << 0);
}

void render()
{   
	// Declare a couple of variables
	int i, j;

	if (isCaptureDone(0))
	{
		int linofs = 0;
		int xofs = 0;
		int yofs = 0;
		int ofs;

        minval = 0;
        scaler = 0xff;
        scaleg = 0xff;
        scaleb = 0xff;
#ifdef FIGHT_AUTO_LIGHT
        int minr=0xff,ming=0xff,minb=0xff;
        int maxr=0,maxg=0,maxb=0;

		for (i = 0; i < 240; i++)
		{
			for (j = 0, ofs = yofs + xofs; j < 320; j++, ofs++, linofs++)
			{
                int b = v(j,i,0);
                int g = v(j,i,8);
                int r = v(j,i,16);
                if (minb > b) minb = b;
                if (ming > g) ming = g;
                if (minr > r) minr = r;
                if (maxb < b) maxb = b;
                if (maxg < g) maxg = g;
                if (maxr < r) maxr = r;
            }
        }

        if (maxr && maxg && maxb)
        {
            minval = (minr << 16) | (ming << 8) | (minb);
            scaler = 0xfe00 / (maxr - minr);
            scaleg = 0xfe00 / (maxg - ming);
            scaleb = 0xfe00 / (maxb - minb);
        }
#endif
		// Draw to gScreen
		for (i = 0; i < 240; i++)
		{
			for (j = 0, ofs = yofs + xofs; j < 320; j++, ofs++, linofs++)
			{
                int b = v(319-j,i,0);
                int g = v(319-j,i,8);
                int r = v(319-j,i,16);
                int v = 0;

                if (((rgb2sv(r,g,b) >> 8) & 0xff) < 0x7f ||
                    ((rgb2sv(r,g,b) >> 0) & 0xff) < 0x4f)
                {
                    v=0;
                }
                else
                {
                    v=0xff;
                }
                
                vbuf[ofs] = (vbuf[ofs] * 1 + v) / 2;

                if (vbuf[ofs] > 0x3f)
                {
                    if ((i^j)&1)
                        ((unsigned int*)gScreen->pixels)[ofs] = 0xff000000;
                    else
                        ((unsigned int*)gScreen->pixels)[ofs] = 0xffffffff;
                }
                else
				    ((unsigned int*)gScreen->pixels)[ofs] = (b << 0) | (g << 8) | (r << 16);
                    
			}
			yofs += gScreen->pitch / 4;
		}
#ifdef BOXFILTER
        int temp[320*240];
        memset(temp,0,sizeof(int)*320*240);
        for (i = 1; i < 239; i++)
            for (j = 1; j < 319; j++)
                temp[i*320+j] = (
                    vbuf[(i-1)*320+j-1] + vbuf[(i-1)*320+j-0] + vbuf[(i-1)*320+j+1] +
                    vbuf[(i-0)*320+j-1] + vbuf[(i-0)*320+j-0] + vbuf[(i-0)*320+j+1] +
                    vbuf[(i+1)*320+j-1] + vbuf[(i+1)*320+j-0] + vbuf[(i+1)*320+j+1]
                                ) / 9;
        memcpy(vbuf,temp,sizeof(int)*320*240);
#endif
        level_l = 239;
        level_r = 239;

        for (i = 239; i > 1; i--)
        {
            int count = 0;
            for (j = 1; j < 159; j++)
            {
                if (vbuf[i*320+j] > 0x3f) count++;
            }
            if (count > 8) level_l = i;
            count = 0;
            for (j = 160; j < 319; j++)
            {
                if (vbuf[i*320+j] > 0x3f) count++;
            }
            if (count > 8) level_r = i;
        }

        for (i = 0; i < 320; i++)
        {
            j = audiolevels[i*4];
            j = ((j * 240) / 0xffff) + 120;
            if (j < 0) j = 0;
            if (j > 239) j = 239;
            *(((unsigned int*)gScreen->pixels)+(gScreen->pitch/4)*j+i)=0xffffffff;

        }

        i = level_l;
        if (i < 1) i = 1;
        if (i > 238) i = 238;
        memset(((unsigned int*)gScreen->pixels)+(gScreen->pitch/4) * (i-1),0,160*4);
        memset(((unsigned int*)gScreen->pixels)+(gScreen->pitch/4) * (i+0),0xff,160*4);
        memset(((unsigned int*)gScreen->pixels)+(gScreen->pitch/4) * (i+1),0,160*4);
        i = level_r;
        if (i < 1) i = 1;
        if (i > 238) i = 238;
        memset(((unsigned int*)gScreen->pixels)+((gScreen->pitch/4) * (i-1))+160,0,160*4);        
        memset(((unsigned int*)gScreen->pixels)+((gScreen->pitch/4) * (i+0))+160,0xff,160*4);        
        memset(((unsigned int*)gScreen->pixels)+((gScreen->pitch/4) * (i+1))+160,0,160*4);        

		doCapture(0);
	}

	Sleep(1);

	// Tell SDL to update the whole gScreen
    SDL_UpdateRect(gScreen, 0, 0, 320, 240);    
}

SDL_AudioSpec *gAudioSpec = NULL;

double sample = 0;
double tone = 0;
double volume = 0;

void audiomixer(void *userdata, Uint8 *stream, int len)
{
    int n = len / 2;
    short * buf = (short*)stream;
    memset(stream,0,len);
    int i;
    for (i = 0; i < n; i++)
    {
        double f = 0;
        int j;
        sample += tone;
        tone   = (((239 - level_r) / 10000.0) + tone * 4000) / 4001;
        volume = (239 - level_l) / 240.0;
        
        f = (sin(sample) * (sample-(floor(sample))-0.5)) * volume;

        audiolevels[i] = buf[i] = (short)(f*0x7fff+0x7ffff); 
    }
}


// Entry point
int main(int argc, char *argv[])
{
	// Initialize SDL's subsystems - in this case, only video.
    if ( SDL_Init(SDL_INIT_VIDEO | SDL_INIT_AUDIO) < 0 ) 
	{
        fprintf(stderr, "Unable to init SDL: %s\n", SDL_GetError());
        exit(1);
    }


	// Register SDL_Quit to be called at exit; makes sure things are
	// cleaned up when we quit.
    atexit(SDL_Quit);

	if (setupESCAPI() < 1)
	{
        fprintf(stderr, "Unable to init ESCAPI\n");
        exit(1);
	}

	gCapture.mWidth = 320;
	gCapture.mHeight = 240;
	gCapture.mTargetBuf = new int[320*240];
	initCapture(0,&gCapture);
	doCapture(0);

    
	// Attempt to create a 640x480 window with 32bit pixels.
    gScreen = SDL_SetVideoMode(320, 240, 32, SDL_SWSURFACE);
	
	// If we fail, return error.
    if ( gScreen == NULL ) 
	{
        fprintf(stderr, "Unable to set 320x240 video: %s\n", SDL_GetError());
        exit(1);
    }
    SDL_AudioSpec *as = new SDL_AudioSpec;
    as->freq = 44100;
    as->format = AUDIO_S16;
    as->channels = 2;
    as->samples = 4096;
    as->callback = audiomixer;
    as->userdata = NULL;
    if (SDL_OpenAudio(as, NULL) < 0)
    {
        fprintf(stderr, "Unable to init SDL audio: %s\n", SDL_GetError());
        exit(1);
    }
    gAudioSpec = as;

    SDL_PauseAudio(0);


	// Main loop: loop forever.
    while (1)
    {
		// Render stuff
        render();

		// Poll for events, and handle the ones we care about.
        SDL_Event event;
        while (SDL_PollEvent(&event)) 
        {
            switch (event.type) 
            {
            case SDL_KEYDOWN:
                break;
            case SDL_KEYUP:
				// If escape is pressed, return (and thus, quit)
                if (event.key.keysym.sym == SDLK_ESCAPE)
				{
					deinitCapture(0);
                    return 0;
				}
                break;
            case SDL_QUIT:
                return(0);
            }
        }
    }
    return 0;
}